/*
 * C45Saver
 * 
 * Created on Apr 5, 2004 By anilkpatro
 */
package weka.core.converters;

import java.io.*;

import weka.core.Attribute;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * C45Saver
 * 
 * @author anilkpatro
 * @since Apr 5, 2004
 */
public class C45Saver extends AbstractSaver
{
    private String  m_fileStem;
    private String  m_path;

    private boolean m_isStructureWritten = false;

    /**
     * Default implementation throws an IOException.
     * 
     * @param file
     *            the File
     * @throws java.io.IOException
     */
    public void setDestination(File file) throws IOException
    {
        if (file == null) {
            throw new IOException("Destination file object is null!");
        }

        String fname = file.getName();
        m_path = file.getParent();
        if (m_path != null) {
            m_path += File.separator;
        } else {
            m_path = "";
        }
        if (fname.indexOf('.') < 0) {
            m_fileStem = fname;
        } else {
            m_fileStem = fname.substring(0, fname.indexOf('.'));
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see weka.core.converters.AbstractSaver#saveStructure(weka.core.Instances)
     */
    public void saveStructure(Instances instances) throws IOException
    {
        BufferedWriter namesWriter;
        String fname = m_path + m_fileStem + ".names";
        File namesFile = new File(fname);
        try {
            namesWriter = new BufferedWriter(new FileWriter(namesFile));
        } catch (FileNotFoundException ex) {
            throw new IOException("File not found : " + (fname));
        }

        try {
            writeNames(namesWriter, instances);
        } catch (Exception e) {
            e.printStackTrace();
        }
        m_isStructureWritten = true;

        namesWriter.close();
    }

    /**
     * @param instances
     */
    private void writeNames(BufferedWriter namesWriter, Instances instances)
            throws Exception
    {
        checkClassAttributes(instances);

        writeAttributeValues(namesWriter, instances, instances.classAttribute());
        for (int i = 0; i < instances.numAttributes(); i++) {
            if (i == instances.classIndex())
                continue;
            Attribute a = instances.attribute(i);
            writeName(namesWriter, a.name());
            namesWriter.write(": ");
            writeAttributeValues(namesWriter, instances, a);
        }
        namesWriter.flush();
    }

    /**
     * @param instances
     */
    private void checkClassAttributes(Instances instances) throws Exception
    {
        if (instances.classIndex() < 0)
            //throw new Exception("Class attribute invalid/undefined");
            instances.setClassIndex(instances.numAttributes() - 1);
        if (!instances.classAttribute().isNominal())
            // Plain vanilla c45 can only handle nominal prediction.
            throw new Exception("Class attribute must be nominal");
        if (instances.classAttribute().numValues() < 2)
            // Quinlan 1993, p. 82: "There must be at least two class names"
            throw new Exception("Not enough classes found");
    }

    protected void writeAttributeValues(BufferedWriter namesWriter,
            Instances instances, Attribute att) throws Exception
    {
        if (att.isNumeric())
            namesWriter.write("continuous");
        else if (att.isString())
            namesWriter.write("discrete " + instances.numInstances());
        else { // att.isNominal()
            if (att.numValues() < 1)
                throw new Exception("Nominal attribute " + att
                        + " has no values.  How did this happen?");
            writeName(namesWriter, att.value(0));
            for (int i = 1; i < att.numValues(); i++) {
                namesWriter.write(", ");
                writeName(namesWriter, att.value(i));
            }
        }
        namesWriter.write(".\n");
    }

    protected void writeName(BufferedWriter writer, String name)
            throws IOException
    {
        char[] chars = name.toCharArray();
        for (int i = 0; i < chars.length; i++) {
            char ch = chars[i];
            switch (ch) {
            case ' ':
                writer.write("\\ ");
                break;
            case ',':
                writer.write("\\,");
                break;
            case '.':
                writer.write("\\.");
                break;
            case ':':
                writer.write("\\:");
                break;
            case '?':
                writer.write("\\?");
                break;
            case '\\':
                writer.write("\\\\");
                break;
            case '|':
                writer.write("\\|");
                break;
            default:
                writer.write(ch);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see weka.core.converters.AbstractSaver#saveDataSet(weka.core.Instances)
     */
    public void saveDataSet(Instances instances) throws Exception
    {
        if (!m_isStructureWritten) {
            saveStructure(instances);
        }

        BufferedWriter dataWriter;
        String fname = m_path + m_fileStem + ".data";
        File dataFile = new File(fname);
        try {
            dataWriter = new BufferedWriter(new FileWriter(dataFile));
        } catch (FileNotFoundException ex) {
            throw new IOException("File not found : " + (fname));
        }

        checkClassAttributes(instances);
        for (int j = 0; j < instances.numInstances(); j++) {
            Instance inst = instances.instance(j);
            int i;
            if (instances.classIndex() == 0) {
                writeInstanceValue(dataWriter, inst, 1);
                i = 2;
            } else {
                writeInstanceValue(dataWriter, inst, 0);
                i = 1;
            }
            for (; i < instances.numAttributes(); i++) {
                if (i == instances.classIndex())
                    continue;
                dataWriter.write(", ");
                writeInstanceValue(dataWriter, inst, i);
            }
            dataWriter.write(", ");
            writeInstanceValue(dataWriter, inst, instances.classIndex());
            dataWriter.write(".\n");
        }
        dataWriter.flush();

        dataWriter.close();
    }

    /**
     * @param inst
     * @param index
     */
    private void writeInstanceValue(BufferedWriter dataWriter, Instance inst,
            int index) throws Exception
    {
        if (inst.isMissing(index))
            dataWriter.write("?");
        else if (inst.attribute(index).isNumeric())
            dataWriter.write(Utils.doubleToString(inst.value(index), 6));
        else
            writeName(dataWriter, inst.stringValue(index));
    }
}